#include "xpack.h"

struct problem P;
struct position *active = NULL;
struct active a = { 0, 0, 0, 0, NULL };
int state = 0;	/* no tile selected */

void free_problem(void) {
    int i, y;
    for (y = 0; y < P.t[0].height+2; ++y)
	free_(P.board[y]);
    free_(P.board);
    for (i = 0; i <= P.maxtile; ++i)
	if (P.t[i].map) {
	    free_(P.t[i].map[0]);
	    free_(P.t[i].map);
	}
    free_(P.t);
    free_(P.position);
    if (a.map) {	
	free_(a.map[0]);
	free_(a.map);
    }
}

void generate_rectangle_problem(int n, int f, int d, int width, int height) {
    int i, sum = 0;
    if (n < 1)
	n = 1;
    P.maxscore = 0;
    P.maxtile = n;
    P.totaltiles = n;
    P.t = malloc_((n+1) * sizeof(struct tile));
    P.position = malloc_((n+1) * sizeof(struct position));
    for (i = 0; i <= n; ++i) {
	P.t[i].height = n+1 - i;
	P.t[i].width = P.t[i].height * f + d;
	P.t[i].type = P.t[i].width == P.t[i].height ? 0 : ALLOW_90;
	P.t[i].count = 1;
	P.t[i].map = NULL;
	if (i) {
	    sum += (P.t[i].score = P.t[i].width * P.t[i].height);
	    P.maxscore += P.t[i].count * P.t[i].score;
	}
	sprintf(P.t[i].name, "%d", n+1-i);
    }	
    if (width && height) {
	P.t[0].width = width;
	P.t[0].height = height;
    } else if (width) {
	P.t[0].width = width;
	P.t[0].height = (sum - 1) / width + 1;
    } else if (height) {
	P.t[0].width = (sum - 1) / height + 1;
	P.t[0].height = height;
    } else {
#if 0
	P.t[0].height = isqrt(sum / f);
	while (P.t[0].height * (P.t[0].height * f + d) < sum)
	    ++P.t[0].height;
	P.t[0].width = P.t[0].height * f + d;
#else
	P.t[0].height = 1 + isqrt(sum - 1);
	P.t[0].width = P.t[0].height;
#endif
    }
    /* printf("Problem uses %d units, square = %d\n",
       sum, P.t[0].width); */
    P.board = malloc_((P.t[0].height+2) * sizeof(int *));
    for (i = 0; i < P.t[0].height+2; ++i)
	P.board[i] = malloc_((P.t[0].width+2) * sizeof(int));
    sprintf(P.description, "Rectangle%02d.%d.%d in %d*%d", n, f, d,
	    P.t[0].width, P.t[0].height);
    a.map = NULL;
}

static char **mapalloc(int w, int h) {
    char **s;
    int i;
    s = malloc_(h * sizeof(char*));
    s[0] = malloc_(w*h);
    memset(s[0], 0, w*h);
    for (i = 1; i < h; ++i)
	s[i] = s[i-1] + w;
    return s;
}

void read_problem(const char *filename) {
    FILE *fp;
    char line[256];
    int i, maxdimen = 0;
    if (!(fp = fopen(filename, "r")))
	fatal("Could not open problem file %s", filename);
    fgets(line, sizeof(line), fp);
    if (strchr(line, '\n'))
	*strchr(line, '\n') = '\0';
    strcpy(P.description, line);
    P.maxscore = 0;
    fscanf(fp, "%d", &P.maxtile);
    P.totaltiles = 0;
    P.t = malloc_((P.maxtile+1) * sizeof(struct tile));
    for (i = 0; i <= P.maxtile; ++i) {
	fscanf(fp, "%d %d %d %d %d %s\n",
	       &P.t[i].count,
	       &P.t[i].type,
	       &P.t[i].width,
	       &P.t[i].height,
	       &P.t[i].score,
	       P.t[i].name);
#if 0
	printf ("read %d %d %d %d %d %s\n",
	       P.t[i].count,
	       P.t[i].type,
	       P.t[i].width,
	       P.t[i].height,
	       P.t[i].score,
	       P.t[i].name);
#endif
	if (P.t[i].type < 1000)
	    P.t[i].map = NULL;
	else {
	    int k;
	    P.t[i].type -= 1000;
	    if (maxdimen < P.t[i].width)
		maxdimen = P.t[i].width;
	    if (maxdimen < P.t[i].height)
		maxdimen = P.t[i].height;
	    P.t[i].map = mapalloc(P.t[i].width, P.t[i].height);
	    for (k = 0; k < P.t[i].height; ++k) {
		char s[80];
		int xx;
		fgets(s, sizeof(s), fp);
		for (xx = 0; xx < P.t[i].width; ++xx) {
		    if (s[xx] == '\n' || !s[xx])
			break;
		    if (s[xx] != ' ' && s[xx] != '.')
			P.t[i].map[k][xx] = 1;
		}
	    }
	}
	
	if (i) {
	    P.maxscore += P.t[i].count * P.t[i].score;
	    P.totaltiles += P.t[i].count;
	}
    }	
    P.position = malloc_((P.totaltiles+1) * sizeof(struct position));
    P.board = malloc_((P.t[0].height+2) * sizeof(int *));
    for (i = 0; i < P.t[0].height+2; ++i)
	P.board[i] = malloc_((P.t[0].width+2) * sizeof(int));
    if (maxdimen)
	a.map = mapalloc(maxdimen, maxdimen);
}

void clear_position(void) {
    int i;
    struct position *pp;
    for (i = 0; i < P.t[0].height+2; ++i)
	memset(P.board[i], 0, (P.t[0].width+2) * sizeof(int));
    pp = P.position;
    for (i = 1; i <= P.maxtile; ++i) {
	int j;
	for (j = 0; j < P.t[i].count; ++j) {
	    ++pp;
	    pp->x = pp->y = -1;
	    pp->type = 0;
	    pp->tilenr = i;
	}
    }
    for (i = 0; i < P.t[0].width+2; ++i)
	P.board[0][i] = P.board[P.t[0].height+1][i] = -1;
    for (i = 1; i <= P.t[0].height; ++i)
	P.board[i][0] = P.board[i][P.t[0].width+1] = -1;
    if (P.t[0].map)
	/* there are further squares which are not allowed */
	for (i = 0; i < P.t[0].height; ++i) {
	    int j;
	    for (j = 0; j < P.t[0].width; ++j)
		if (!P.t[0].map[i][j])
		    P.board[i+1][j+1] = -1;
	}
}

int place_active(int x, int y) {
    int i, j;
    if (x < 0 || y < 0 ||
	x + a.w > P.t[0].width ||
	y + a.h > P.t[0].height)
	return 0;
    /* TODO: check type, build active pattern otherwhere */
    if (a.map) {
	for (i = 0; i < a.h; ++i)
	    for (j = 0; j < a.w; ++j)
		if (a.map[i][j] && P.board[y+1+i][x+1+j])
		    return 0;
    } else {	/* rectangle */
	for (i = 0; i < a.h; ++i)
	    for (j = 0; j < a.w; ++j)
		if (P.board[y+1+i][x+1+j])
		    return 0;
    }
    /* placement is OK, really place the tile */
    active->x = x;
    active->y = y;
    active->type = a.type;
    /* different if map TODO */
    {   int k;
    k = active - P.position;
    if (a.map) {
	for (i = 0; i < a.h; ++i)
	    for (j = 0; j < a.w; ++j)
		if (a.map[i][j])
		    P.board[y+1+i][x+1+j] = k;
    } else {
	for (i = 0; i < a.h; ++i)
	    for (j = 0; j < a.w; ++j)
		P.board[y+1+i][x+1+j] = k;
    }
    }
    active = NULL;
    return 1;
}

void discard_active(void) {
    active->x = active->y = -1;
    active->type = 0;
    active = NULL;
}


void createmap(void) {
    int i;
    /* create the new bitmap for the active tile */
    /* printf("Creating map for type %d\n", a.type); */
    if (!a.map)
	return;	/* nothing to do */
    if (!P.t[a.tilenr].map) {
	/* copy rectangle */
	if (a.type & ALLOW_90) {
	    for (i = 0; i < P.t[a.tilenr].width; ++i)
		memset(a.map[i], 1, P.t[a.tilenr].height);
	} else {
	    for (i = 0; i < P.t[a.tilenr].height; ++i)
		memset(a.map[i], 1, P.t[a.tilenr].width);
	}
	return;
    }
    /* make pattern */
    {   int x, y, ix, iy;
	static struct dtab {
	    int dxx, dxy, dyx, dyy;
	    int sx, sy;
	} dtab[] = {
	    {  1,  0,  0,  1,   0, 0 },
	    {  0,  1, -1,  0,   1, 0 },
	    { -1,  0,  0, -1,   1, 2 },
	    {  0, -1,  1,  0,   0, 2 },
	    { -1,  0,  0,  1,   1, 0 },
	    {  0,  1,  1,  0,   0, 0 },
	    {  1,  0,  0, -1,   0, 2 },
	    {  0, -1, -1,  0,   1, 2 } };
	char **omap = P.t[a.tilenr].map;
	int starts[3];
	starts[0] = 0;
	starts[1] = P.t[a.tilenr].width-1;
	starts[2] = P.t[a.tilenr].height-1;
	/* printf("4 = %d, 5 = %d\n", dtab[a.type].sx, dtab[a.type].sy); */
	for (y = 0; y < a.h; ++y) {
	    ix = starts[dtab[a.type].sx];
	    ix += y * dtab[a.type].dyx;
	    iy = starts[dtab[a.type].sy];
	    iy += y * dtab[a.type].dyy;
	    for (x = 0; x < a.w; ++x) {
		a.map[y][x] = omap[iy][ix];
		/* printf("(%d,%d) from (%d,%d)\n", x, y, ix, iy); */
		ix += dtab[a.type].dxx;
		iy += dtab[a.type].dxy;
	    }
	}
    }
}

void activate_free(int k) {
    int tilenr;
    active = P.position + k;
    tilenr = active->tilenr;
    a.w = P.t[tilenr].width;
    a.h = P.t[tilenr].height;
    a.tilenr = tilenr;
    a.type = 0;
    createmap();
}

void activate_used(int k) {
    int i, j, x, y;
    a.tilenr = P.position[k].tilenr;
    active = P.position + k;
    a.type = active->type;
    if (a.type & 1) {
	a.w = P.t[a.tilenr].height;
	a.h = P.t[a.tilenr].width;
    } else {
	a.h = P.t[a.tilenr].height;
	a.w = P.t[a.tilenr].width;
    }
    createmap();
    /* free board */
    x = P.position[k].x + 1;
    y = P.position[k].y + 1;
    if (a.map) {
	for (i = 0; i < a.h; ++i)
	    for (j = 0; j < a.w; ++j)
		if (a.map[i][j])
		    P.board[y+i][j+x] = 0;
    } else {
	for (i = 0; i < a.h; ++i)
	    for (j = 0; j < a.w; ++j)
		P.board[y+i][j+x] = 0;
    }
}

void make_simple_solution(void) {
    int i;
    clear_position();
    for (i = 1; i <= P.totaltiles; ++i) {
	int x, y;
	activate_free(i);
	for (y = 0; y <= P.t[0].height - a.h; ++y)
	    for (x = 0; x <= P.t[0].width - a.w; ++x)
		if (place_active(x, y))
		   goto done;
	discard_active();
    done:
	;
    }
}

void set_type(int i) {
    if ((a.type ^ i) & 1) {
	int x;
	x = a.w;
	a.w = a.h;
	a.h = x;
    }
    a.type = i;
    if ((a.type & ALLOW_180) && !(P.t[a.tilenr].type & ALLOW_180))
	a.type &= ~ALLOW_180;	/* normalise tile */
    createmap();
}
